#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <netdb.h>
#include <dirent.h>

// 
// WARNING: strerror() is not MT safe
// 

#include "netproto.h"
#include "xalloc.h"
#include "server_common.h"
#include "sure_sock.h"

#define DECLARE_get_type(the_type) 					\
static inline int get_##the_type(int sock, the_type *valp) {		\
	int ret;							\
	ret = sure_recv(sock, valp, sizeof(*valp), MSG_NOSIGNAL);	\
	return ret != sizeof(*valp);					\
}

DECLARE_get_type(int16_t)
DECLARE_get_type(int32_t)
DECLARE_get_type(uint16_t)
DECLARE_get_type(uint32_t)

#define DECLARE_put_type(the_type)                                      \
static inline int put_##the_type(int sock, the_type *valp) {            \
        int ret;                                                        \
        ret = sure_send(sock, valp, sizeof(*valp), MSG_NOSIGNAL);       \
        return ret != sizeof(*valp);                                    \
}

DECLARE_put_type(int16_t)
DECLARE_put_type(int32_t)
DECLARE_put_type(uint16_t)
DECLARE_put_type(uint32_t)



static int
compare_files(const void *obj_a, const void *obj_b)
{
	const struct valid_file_t *a = (const struct valid_file_t *) obj_a;
	const struct valid_file_t *b = (const struct valid_file_t *) obj_b;

	if (a->inode < b->inode) {
		return -1;
	} else if (a->inode > b->inode) {
		return 1;
	}
	return 0;
}

static void
get_server_files(int sock, struct valid_file_t **array, int *array_size)
{
	uint32_t magic;
	uint16_t size;
	//uint32_t checksum;
	uint32_t ack;
	char *datap;
	int ret;

	*array_size = -1;

	if (get_uint32_t(sock, &magic)) {
		fprintf(stderr, "ERROR: cannot get_uint16_t() errno=%d\n",errno);
		return;
	}

	if (magic != SCANDISK_NET_MAGIC) {
		fprintf(stderr, "ERROR: not enough magic 0x%04x instead of 0x%04x\n", magic, SCANDISK_NET_MAGIC);
		return;
	}
	if (get_uint16_t(sock, &size)) {
		fprintf(stderr, "ERROR: cannot get_uint16_t() errno=%d\n", errno);
		return;
	}
	if (size == 0) {
		fprintf(stderr, "ERROR: size of data to receive is 0\n");
		return;
	}
	if (size % sizeof(struct valid_file_t)) {
		fprintf(stderr, "ERROR: size does not allow an integer number of records\n");
		return;
	}

	//if (get_uint16_t(sock, &checksum)) {
	//}

	datap = malloc(size);
	if (!datap) {
		fprintf(stderr, "ERROR: requesting too much size (%d bytes)\n", size);
		return;
	}
	ret = sure_recv(sock, datap, size, MSG_NOSIGNAL);
	if (ret == -1) {
		xfree(datap);
		return;
	} else if (ret == 0) {
		xfree(datap);
		return;
	}

	ack = SCANDISK_NET_MAGIC_ACK;
	if (put_uint32_t(sock, &ack)) {
		fprintf(stderr, "ERROR: cannot put_uint32_t() errno=%d\n", errno);
		return;
	}

	*array = (struct valid_file_t *) datap;
	*array_size = size / sizeof (struct valid_file_t);

}

void do_parse_dir(char *datapath, const struct valid_file_t * const file_array, const int file_array_len) {
#define NB_IODS_SUBDIR 0x20

// path/file looks like XX/iXXXXXXXXsXXXXXXXX
#define KEYMAXSIZE (2+1+((1+8)*2)+1)
	int nsubdir;
	for (nsubdir = 0; nsubdir < NB_IODS_SUBDIR ; nsubdir++) {
		char *currentdir;
		int nfiles;
		struct dirent **filelist;

		// currentdir is datapath + "/XX/iXXXXXXXXsXXXXXXXX" + final '\0'
//		currentdir = xmalloc(strlen(datapath) + strlen("/XX/iXXXXXXXXsXXXXXXXX") + 1);
		currentdir = xmalloc(strlen(datapath) + strlen("/XX/") + 1);

		sprintf(currentdir,"%s/%02x/", datapath, nsubdir );
		printf("Scanning %s:\n", currentdir);
		
		nfiles = scandir(currentdir, &filelist, NULL, NULL);
		if (nfiles == -1) {
			goto out_error;
			fprintf(stderr, "ERROR: an error occurred in scandir of %s\n", currentdir);
		} else {
			while (nfiles--) { 
				char *inode_str, *seed_str, *filename;
				int filename_len;
				uint32_t inode, seed;

				filename = filelist[nfiles]->d_name;
				filename_len = strlen(filename);

				if (filename_len == strlen("iXXXXXXXXsXXXXXXXX")) {
					filename[0]=0;
					filename[9]=0;
					
					inode_str = filename + 1;
					seed_str = filename + 10;

					inode = strtol(inode_str, NULL, 16);
					seed = strtol(seed_str, NULL, 16);

					struct valid_file_t vf;
					vf.inode = inode;
					vf.seed = seed;


					if (bsearch(&vf, file_array, file_array_len,
							sizeof(vf), compare_files)) {
						printf("[%s][%s] FOUND -> linked\n",
								inode_str, seed_str);
					} else {
						printf("[%s][%s] NOT FOUND -> not linked\n",
								inode_str, seed_str);
					}
				} else {
					printf("%s -> BLEH\n", filelist[nfiles]->d_name);
				}
				free(filelist[nfiles]);
			}
			free(filelist);
		}
out_error:
		xfree(currentdir);
	}
}

static int do_continue = 1;

void sighdl(int i) {
	do_continue = 0;
}


/*
 * The algorithm used is fairly simple:
 *  - wait a connection
 *  - receives the valid couples (inode,seed) from the nfsd server
 *  - close the connection
 *  - check for each file in /datapath/??/i????????s???????? if it has a matching
 *    (inode,seed) couple and hence still linked
 *
 * The problem with this approach is that we cannot do an online scandisk
 * (without stopping first the nfsp servers) as some files may have been
 * created on the server and its data stored, before the full list of couples
 * is transfered to us...
 *
 * Hence scandisks must be done offline.
 *
 * To be exactly synchronous, the iods must be put in drop mode then the nfspd
 * then the scandisk done then the iods resumed then the nfspd.
 *  
 */

int
main(int argc, char *argv[])
{
	char *datapath;
	int i;
	int serversock;
	int sock;
	struct sockaddr_in client_in;
	socklen_t client_in_size;

	struct valid_file_t *linked_files;
	int linked_files_len;

	if (argc != 2) {
		fprintf(stderr, "ERROR: %s /path/to/storage\n", argv[0]);
		exit(1);
	}

	datapath = strdup(argv[1]);
	if (access(datapath, R_OK|W_OK|X_OK)) {
		fprintf(stderr, "ERROR: %s is not accessible in read+write+exec mode\n",
				datapath);
		exit(1);
	}
	printf("The datapath [%s] seems OK.", datapath);

	serversock = setserversock("20222", "tcp", 5);

	while (do_continue) {
		client_in_size = sizeof(client_in);
		sock = accept(serversock, (struct sockaddr *) (&client_in), 
				&client_in_size);
		if (sock == -1) {
			switch (errno) {
			case EBADF:
			case ENOTSOCK:
			case ENOTSUP:
			case EFAULT:
			case ENOMEM:
				fprintf(stderr,
					"ERROR: exiting accept() errno=%d [%s]",
					errno, strerror(errno));
				exit(1);
			default:
				fprintf(stderr,
					"ERROR: accept() returned errno=%d [%s]",
					errno, strerror(errno));
			}
		}

		get_server_files(sock, &linked_files, &linked_files_len);
		if (linked_files_len < 0) {
			fprintf(stderr, "ERROR: occured in get_server_files()\n");
			goto close_conn;
		}

		qsort(linked_files, linked_files_len, sizeof (*linked_files),
		      compare_files);


		printf("%d files\n", linked_files_len);
		printf("inode\tseed\n");
		for (i = 0; i < linked_files_len; i++) {
			printf("%8lx\t%8x\n",
			       linked_files[i].inode,
			       linked_files[i].seed);
		}

		do_parse_dir(datapath, linked_files, linked_files_len);

close_conn:
		int ret = close(sock);
		if (ret == -1) {
			fprintf(stderr, "ERROR: close() returned an error errno=%d\n", errno);
		}
	}


	return 0;
}
